import net.jxta.platform.NetworkManager;
import net.jxta.platform.NetworkConfigurator;

import net.jxta.peergroup.PeerGroup;
import net.jxta.peergroup.PeerGroupID;

import net.jxta.discovery.DiscoveryService;
import net.jxta.discovery.DiscoveryListener;
import net.jxta.discovery.DiscoveryEvent;

import net.jxta.protocol.DiscoveryResponseMsg;
import net.jxta.protocol.PeerAdvertisement;
import net.jxta.protocol.PipeAdvertisement;

import net.jxta.id.IDFactory;

import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Advertisement;
import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredDocumentUtils;
import net.jxta.document.StructuredTextDocument;

import net.jxta.pipe.PipeService;
import net.jxta.pipe.PipeID;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.OutputPipeEvent;
import net.jxta.pipe.OutputPipeListener;
import net.jxta.pipe.PipeMsgEvent;
import net.jxta.pipe.PipeMsgListener;

import net.jxta.endpoint.Message;
import net.jxta.endpoint.Message.ElementIterator;
import net.jxta.endpoint.MessageElement;
import net.jxta.endpoint.StringMessageElement;
import net.jxta.endpoint.WireFormatMessage;
import net.jxta.endpoint.WireFormatMessageFactory;

import net.jxta.util.CountingOutputStream;
import net.jxta.util.DevNullOutputStream;

import java.io.File;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Scanner;

import java.net.URI;
import java.net.URISyntaxException;



public class Client implements DiscoveryListener, OutputPipeListener, PipeMsgListener{
	/* attributes */
//	private final Object lock = new String("lock");

	private NetworkManager manager;
	private NetworkConfigurator theNetworkConfigurator;
	private PeerGroup thePeerGroup;
	private DiscoveryService discovery;
	
	private PipeAdvertisement myPipeAdv, hisPipeAdv;
	private PeerAdvertisement friendAdvertisement;

	private boolean friendDiscovered = false;
	private boolean friendPipeDiscovered = false;
	private boolean pipeBindingOK = false;
	
	private int protocolPhase = 0;

	private PipeService pipeService;
	private OutputPipe outputPipe;
	private InputPipe inputPipe;

	private static boolean verbosity;
	/**
	* Constructor for the Client. Fetches the netPeerGroup and the discovery
	* service associated with it. Initializes the pipe service
	* @param[in] clientName The local client name, that is your ID
	*/
	public Client(String clientName) {
		/* initialize the network manager */
		try{
			manager = new NetworkManager( NetworkManager.ConfigMode.ADHOC, "Chat", new File( new File(".cache"), clientName ).toURI() );  
			manager.setConfigPersistent(false);
			theNetworkConfigurator = manager.getConfigurator();
			
			if (theNetworkConfigurator.exists()){
				System.out.println(">INFO: Configuration found ...");
				File networkConfig = new File ( theNetworkConfigurator.getHome(), "PlatformConfig" );
				try{
					theNetworkConfigurator.load( networkConfig.toURI());
				}catch(Exception e){
					e.printStackTrace();
					System.exit(-1);
				}
			}else{
				System.out.println(">INFO: Configuration not found. Creating new one !");
				theNetworkConfigurator.setName( clientName);
				theNetworkConfigurator.setPrincipal ( clientName );
				theNetworkConfigurator.setPassword ("12345678");
				theNetworkConfigurator.save();
			}
			
			System.out.println(">INFO: Starting JXTA");
			thePeerGroup = manager.startNetwork();
			System.out.println(">INFO: JXTA Started");
			
			discovery = thePeerGroup.getDiscoveryService();
			pipeService = manager.getNetPeerGroup().getPipeService();
		}catch(Exception e){
			e.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	* main
	* @param args command line args
	*/
	public static void main(String args[]) {
		//check the correct number of arguments
		if ( args.length != 3 ) {
			System.out.println(">ERROR: usage java Client id1 id2 verb");
			System.out.println(">             id1 : your ID ");
			System.out.println(">             id2 : his ID  ");
			System.out.println(">             verb: verebosity 0/1");
			System.out.println(">EXAMPLE: java Client alice bill 0 2> /dev/null");
			System.exit(-1);	
		}
		
		int iverb = Integer.parseInt(args[2]);
		if ( iverb != 0 && iverb != 1){
			System.out.println("ERROR: Third argument should be 0 or 1");
			System.exit(-1);
		}else{
			if (iverb == 0)
				verbosity = false;
			else
				verbosity = true;
		}
		System.out.println(">INFO: Verbosity was set to " + verbosity);
					
		Client chatClient = new Client( args[0]);     /* initialize the jxta network, discovery and pipe service */
		chatClient.initLocalCache(verbosity);         /* clear local advertisement cache otherwise conflicts arise */
		chatClient.getClientInfo();                   /* print client info */ 
		chatClient.discoverPartenerClient(args[1]);   /* discover the chat partener having the name given by args[1] */
		System.out.println(">INFO: Found matching peer on the JXTA network ");
		/* create a pipeAdvertisement having the name myID_hisID. This advertisement will be binded to the output pipe*/
		chatClient.myPipeAdv = chatClient.getPipeAdvertisement(args[0]+"_"+args[1]);  			
		System.out.println(">INFO: Succesfully created unicast pipe advertisement having name " + chatClient.myPipeAdv.getName() );
		chatClient.publishPipeAdvertisements(); /* publish the newly created pipe advertisement in the JXTA network */
		System.out.println(">INFO: Published the created pipe advertisement");
		/* fetch from the network the partener pipe advertisement having the name hisID_myID */		
		chatClient.getHisPipeAdvertisement(args[1]+"_"+args[0]);
		System.out.println(">INFO: Partener pipe advertisement having name " + chatClient.hisPipeAdv.getName() + " was successfully fetched" );		
		/* try to bind my pipe advertisement to an output pipe and the partener pipe advertisement to an input pipe */
		chatClient.bindPipeAdvertisements();		
		System.out.println(">INFO: Pipes have been binded successfully");
		/* prints the nice chat header */
		chatClient.printChatHeader();
		chatClient.chat(args[0]); /* start chatting */
		chatClient.stop(); /* stop the JXTA network */
	}

	/**
	* initization of the local advertisement cache. All pipe advertisements are flushed 
	* @param verb print messy details on the scrren 
	*/
	public void initLocalCache(boolean verb){
		Advertisement adv;
		System.out.println(">INFO: Printing existing local advertisements");
		try{
			Enumeration en = discovery.getLocalAdvertisements(DiscoveryService.ADV, null, null);
			if (en != null) {
				while (en.hasMoreElements()) {
					adv = (Advertisement) en.nextElement();
					if ( adv instanceof PipeAdvertisement ){
						if (verb) 
							System.out.println(">INFO: --------- Pipe Advertisement ---------");
					}
					if (verb){
						System.out.println(">INFO: ------------------------------------");					
						System.out.println(adv.toString());
					}
					if ( adv instanceof PipeAdvertisement ){
						if (verb)
							System.out.println(">INFO: .... flushing pipe advertisement ............"	);
						discovery.flushAdvertisement(adv);
					}
				}
			}
		}catch(Exception e){
			e.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	* Prints the client details: name, ID, group name, group ID
	*/
	void getClientInfo(){
		System.out.println ( ">INFO: Peer name       :" + thePeerGroup.getPeerName() );
		System.out.println ( ">INFO: Peer ID         :" + thePeerGroup.getPeerID().toString() );
		System.out.println ( ">INFO: Peer group name :" + thePeerGroup.getPeerGroupName() );
		System.out.println ( ">INFO: Peer group ID   :" + thePeerGroup.getPeerGroupID().toString() );
	}

	/**
	* loop attempting to discover advertisements the partener peer advertisement
	* @param[in] partenerName the name of the peer we wish to discover
	*/
	public void discoverPartenerClient(String partenerName) {
		long waittime = 6 * 1000L;
		try {
			while (!friendDiscovered) {
				System.out.println(">INFO: Sending a Friend Discovery Message");
				discovery.getRemoteAdvertisements(
				        null,                  // no specific peer (propagate)
				        DiscoveryService.PEER, // Adv type one of DiscoveryService.PEER, DiscoveryService.GROUP, DiscoveryService.ADV
				        "Name",                // Attribute = attribute name to narrow discovery to
				        partenerName,          // Value =  value of attribute to narrow discovery to
				        1,                     // the upper limit of responses from one peer
				        this                   // no query specific listener. we are using a global listener
						);
				try {
					System.out.println(">INFO: Sleeping for :" + waittime);
					Thread.sleep(waittime);
				} catch (Exception e) {
					// ignored
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	* Creates a pipe advertisement with a given name
	* @param[in] pipeName the name of the pipe advertisement to be created
	* @return a Pipe Advertisement
	*/
	public PipeAdvertisement getPipeAdvertisement(String pipeName) {
		PipeAdvertisement advertisement = (PipeAdvertisement) AdvertisementFactory.newAdvertisement(PipeAdvertisement.getAdvertisementType());
		advertisement.setPipeID(IDFactory.newPipeID(PeerGroupID.defaultNetPeerGroupID));
		advertisement.setType(PipeService.UnicastType);
		advertisement.setName(pipeName);
		return advertisement;
	}

	/**
	* Published my pipe advertisement both locally and remotely 
	*/
	public void publishPipeAdvertisements(){
		try{
			discovery.publish(myPipeAdv);
			discovery.remotePublish(myPipeAdv);
		}catch(Exception e){
			System.out.println(">ERROR: Problem publishing advertisements");
			System.exit(-1);
		}
	}
	
	/**
	* Send a discovery request for the partener pipe advertisement having the given name
	* @param[in] pipeName the name of the pipe advertisement be searched
	*/
	public void getHisPipeAdvertisement(String partenerPipeName) {
		long waittime = 6 * 1000L;
		System.out.println(">INFO: Trying to recover pipe advertisement having name: " + partenerPipeName);
		try {
			while (!friendPipeDiscovered) {
				System.out.println(">INFO: Sending a Friend Pipe Discovery Message");
				discovery.getRemoteAdvertisements(
				        null,                  // no specific peer (propagate)
				        DiscoveryService.ADV, // Adv type one of DiscoveryService.PEER, DiscoveryService.GROUP, DiscoveryService.ADV
				        "Name",                // Attribute = attribute name to narrow discovery to
				        partenerPipeName,      // Value =  value of attribute to narrow discovery to
				        1,                     // the upper limit of responses from one peer
				        this                   // no query specific listener. we are using a global listener
						);
				try {
					System.out.println(">INFO: Sleeping for :" + waittime);
					Thread.sleep(waittime);
				} catch (Exception e) {
				// ignored
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	* Tries to bind my pipe advertisement to the output pipe and the partener
	* pipe advertisement to the input pipe
	*/
	public void bindPipeAdvertisements(){
		try{
			System.out.println(">INFO: Binding input pipe on advertisement " + hisPipeAdv.getName());
			inputPipe = pipeService.createInputPipe(hisPipeAdv, this);
			while (! pipeBindingOK){
				try {
					Thread.sleep( 3000);
				} catch (Exception e) {
					// ignored
				}
				System.out.println(">INFO: Trying to bind output pipe to advertisement " + myPipeAdv.getName());
				pipeService.createOutputPipe(myPipeAdv, this);
			}
		}catch( Exception e){
			System.out.println(">ERROR: problem binding pipes");
			System.exit(-1);
		}
	}

	/**
	* This is where my messages are read from the keyboard and sent out the output pipe
	* @param[in] myID my ID in the chat
	*/
	public void chat(String myID){
		String theChatLine; /*just a string that will contain my lines of text */
		/* a scanner of the System.in. Various methods can be applied from fetching
		  entire lines to fetching integers */
		Scanner in = new Scanner(System.in); 
		
		/* loop forever. Note that this is not a good design practice. We should
		   loop function of a boolean parameter */
		while(true){
			/* Reads a single line from the console and stores into name variable*/
			theChatLine = in.nextLine();
			/* check if we tipped "exit". If yes we leave the chat */
			if (theChatLine.compareTo("exit") == 0) {
				in.close();  /* close the scanner */
				inputPipe.close();
				outputPipe.close();
				try{ /*try to flush the local cache of advertisements */
					discovery.flushAdvertisement(myPipeAdv);
					discovery.flushAdvertisement(hisPipeAdv);
					discovery.flushAdvertisement(friendAdvertisement);
				}catch(Exception e){}				
				stop(); /*close jxta */
				System.out.println(">INFO: Bye Bye ...");
				System.exit(1);
			}
			/* print my message on the screen */
			System.out.println("[ " + myID + " ]: " + theChatLine );
			/* send the message on the output pipe */
			Message msg;
			try {
				if (verbosity)
					System.out.println(">INFO: Sending message");
				/* create the message */
				msg = new Message();
				/* message is formed out of message elements, so we create a new message 
				   element with the name "chat tutorial" and containing the line of chat */
				StringMessageElement sme = new StringMessageElement("Chat tutorial", theChatLine, null);
				/* we add this messge element to our message */
				msg.addMessageElement(null, sme);
				/* we send out the message */
				outputPipe.send(msg);
				if (verbosity)
					System.out.println(">INFO: message sent");
			} catch (IOException e) {
					System.out.println(">ERROR: failed to send message");
					e.printStackTrace();
					System.exit(-1);
			}
		}
	}

	/**
	* This method is called whenever a discovery response is received, which are
	* either in response to a query we sent, or a remote publish by another node
	* @param ev the discovery event
	*/
	public void discoveryEvent(DiscoveryEvent ev) {
		DiscoveryResponseMsg res = ev.getResponse();
		if (verbosity)
			System.out.println(" [  Got a Discovery Response [" + res.getResponseCount() + " elements]  from peer : " + ev.getSource() + "  ]");

		/* in protocol phase 0 we are ready to receive the partener peer advertisement */
		if (protocolPhase == 0){
			if (verbosity)
				System.out.println(">INFO: Entering phase 0");
			/* I'm expecting the peer advertisement at phase 0 */
			Enumeration en = res.getAdvertisements();
			if (en != null) {
				while (en.hasMoreElements()) {
					friendAdvertisement = (PeerAdvertisement) en.nextElement();
					if (verbosity)
						System.out.println(">INFO: Peer:" + friendAdvertisement.getName());
				}
			}
			if (verbosity)
				System.out.println(">INFO: Printed-out all the advertisments received");
			
			friendDiscovered = true; /* we have discovered our partener in the jxta network */
			protocolPhase = 1;       /* we pass to phase 1 */

		/* in phase 1 i'm expecting the partener pipe advertisement */
		}else if (protocolPhase == 1){
			if (verbosity)
				System.out.println(">INFO: Entering phase 1");
			/* I'm expecting the pipe advertisement at phase 1 */
			Enumeration en = res.getAdvertisements();
			if (en != null) {
				while (en.hasMoreElements()) {
					hisPipeAdv = (PipeAdvertisement) en.nextElement();
					if (verbosity)
						System.out.println(">INFO: Pipe:" + hisPipeAdv.getName());	
				}
			}
			if (verbosity)
				System.out.println(">INFO: Printed-out all the advertisments received");
			
			friendPipeDiscovered = true; /* we have discovered the partenere pipe advertisement */
			protocolPhase = 2;           /* we pass to stage 2 ...nothing should happen there */
		}else {
			/* protocol phase = 2  so we start transmitting messages :) */
			if (verbosity)
				System.out.println(">INFO: Entering phase 2, where we actuall receive messages");
		}
	}

	/**
	* by implementing OutputPipeListener we must define this method which
	* is called when the output pipe tries to be bind to the advertisement.
	* In fact, the output pipe object cannot be created until the input
	* pipe having the same advertisement has been found 
	*
	* @param event event object from which to get output pipe object
	*/
	public void outputPipeEvent(OutputPipeEvent event) {
		if (verbosity)
			System.out.println("INFO: Received the output pipe resolution event");
		outputPipe = event.getOutputPipe();
		pipeBindingOK=true; /* pipe has been successfully binded */
	}

	/**
	* PipeMsgListener interface for asynchronous message arrival notification.
	* Each time a message is received on the input pipe this method is 
	* asynchronously called
	* @param event the message event
	*/
	public void pipeMsgEvent(PipeMsgEvent event) {
		Message msg;
		try {
			/* Obtain the message from the event */
			msg = event.getMessage();
			if (msg == null) {
				System.out.println(">INFO: Received an empty message");
			return;
			}
			printMessageStats(msg, true); /* we dump the contents of the message on the screen */
		} catch (Exception e) {
			e.printStackTrace();
		return;
		}
	}


	/**
	* Dumps the message content to stdout
	*
	* @param msg     the message
	* @param verbose dumps message element content if true
	*/
	public static void printMessageStats(Message msg, boolean verbose) {
		try {
			CountingOutputStream cnt;
			ElementIterator it = msg.getMessageElements();
			System.out.println("------------------Begin Message---------------------");
			WireFormatMessage serialed = WireFormatMessageFactory.toWire(
			msg,
			new MimeMediaType("application/x-jxta-msg"), null);
			System.out.println("Message Size :" + serialed.getByteLength());
			while (it.hasNext()) {
				MessageElement el = it.next();
				String eName = el.getElementName();
				cnt = new CountingOutputStream(new DevNullOutputStream());
				el.sendToStream(cnt);
				long size = cnt.getBytesWritten();
				System.out.println("Element " + eName + " : " + size);
				if (verbose) {
					System.out.println("[" + el + "]");
				}
			}
			System.out.println("-------------------End Message----------------------");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Prints a nice chat header
	 */
	public void printChatHeader(){
		System.out.println("--------------------------------------------------------------------------------");
		System.out.println("-2009 Simple JXTA Chat Demo ----------------------------------------------------");
		System.out.println("-INSTRUCTIONS:  ");
		System.out.println("              to CHAT write your line and press enter ");
		System.out.println("              to LEAVE write \"exit\" (without quotes) and  press enter ");
	}

	/**
		* Stops the platform
		*/
	public void stop() {
		// Stop JXTA
		manager.stopNetwork();
	}
}

